"""Summary statistics (observations and forecasts)."""

import numpy as np


# Add new metrics to this map to map shorthand to function
_MAP = {
    'mean': (np.mean, 'Mean'),
    'min': (np.min, 'Min'),
    'max': (np.max, 'Max'),
    'std': (np.std, 'Std'),
    'median': (np.median, 'Median'),
    'var': (np.var, 'Variance'),
}

__all__ = [m[0].__name__ for m in _MAP.values()]



import calendar
import datetime
import itertools
import re

import numpy as np
from numpy.testing import assert_allclose
import pandas as pd
import pytest


from solarforecastarbiter import datamodel
from solarforecastarbiter.metrics import (calculator, deterministic,
                                          probabilistic, event)


DETERMINISTIC_METRICS = list(deterministic._MAP.keys())
DET_NO_NORM = (set(DETERMINISTIC_METRICS) - set(deterministic._REQ_NORM))
DET_NO_REF = (set(DETERMINISTIC_METRICS) - set(deterministic._REQ_REF_FX))
DET_NO_REF_NO_NORM = (set(DET_NO_NORM) - set(deterministic._REQ_REF_FX))
PROBABILISTIC_METRICS = list(probabilistic._MAP.keys())
# PROB_NO_NORM = (set(PROBABILISTIC_METRICS) - set(probabilistic._REQ_NORM))
PROB_NO_REF = (set(PROBABILISTIC_METRICS) - set(probabilistic._REQ_REF_FX))
PROB_NO_DIST = (set(PROBABILISTIC_METRICS) - set(probabilistic._REQ_DIST))
PROB_NO_REF_NO_DIST = (set(PROB_NO_REF) - set(probabilistic._REQ_DIST))
LIST_OF_CATEGORIES = list(datamodel.ALLOWED_CATEGORIES.keys())
EVENT_METRICS = list(event._MAP.keys())


@pytest.fixture()
def create_processed_fxobs(create_dt_index, banded_cost_params):
    def _create_processed_fxobs(fxobs, fx_values, obs_values,
                                ref_values=None,
                                interval_label=None):

        if not interval_label:
            interval_label = fxobs.forecast.interval_label

        if (isinstance(fx_values, pd.Series) or
           isinstance(fx_values, pd.DataFrame)):
            conv_fx_values = fx_values
        else:
            conv_fx_values = pd.Series(
                fx_values, index=create_dt_index(len(fx_values)))
        if isinstance(obs_values, pd.Series):
            conv_obs_values = obs_values
        else:
            conv_obs_values = pd.Series(
                obs_values, index=create_dt_index(len(obs_values)))

        return datamodel.ProcessedForecastObservation(
            fxobs.forecast.name,
            fxobs,
            fxobs.forecast.interval_value_type,
            fxobs.forecast.interval_length,
            interval_label,
            valid_point_count=len(fx_values),
            forecast_values=conv_fx_values,
            observation_values=conv_obs_values,
            reference_forecast_values=ref_values,
            normalization_factor=fxobs.normalization,
            uncertainty=fxobs.uncertainty,
            cost=banded_cost_params
            )

    return _create_processed_fxobs


@pytest.fixture()
def create_dt_index():
    def _create_dt_index(n_periods):
        return pd.date_range(start='20190801', periods=n_periods, freq='1h',
                             tz='MST', name='timestamp')
    return _create_dt_index


@pytest.fixture()
def copy_prob_forecast_with_axis():
    def _copy_prob_forecast_with_axis(probfx, axis, constant_values=None):
        if constant_values:
            cvs = [probfx.constant_values[0].replace(constant_value=cv)
                   for cv in constant_values]
        else:
            cvs = probfx.constant_values
        new_cvs = []
        new_probfx = probfx.replace(constant_values=())
        new_probfx = new_probfx.replace(axis=axis)
        for cv in cvs:
            cv = cv.replace(axis=axis)
            new_cvs.append(cv)
        new_probfx = new_probfx.replace(constant_values=tuple(new_cvs))
        return new_probfx
    return _copy_prob_forecast_with_axis


@pytest.fixture(params=['fxobs', 'fxagg', 'fxobs_ref', 'fxagg_ref'])
def single_forecast_data_obj(
        request, single_forecast_aggregate,
        single_forecast_observation,
        single_forecast_aggregate_reffx,
        single_forecast_observation_reffx):
    if request.param == 'fxobs':
        return single_forecast_observation
    elif request.param == 'fxagg':
        return single_forecast_aggregate
    if request.param == 'fxobs_ref':
        return single_forecast_observation_reffx
    elif request.param == 'fxagg_ref':
        return single_forecast_aggregate_reffx


@pytest.fixture(params=['probfxobs', 'probfxagg',
                        'probfx_ref', 'probfxagg_ref', 'probfxobsy'])
def prob_forecasts_data_obj(
        request, single_prob_forecast_observation,
        single_prob_forecast_aggregate,
        single_prob_forecast_observation_y,
        single_prob_forecast_observation_reffx,
        single_prob_forecast_aggregate_reffx):
    if request.param == 'probfxobs':
        return single_prob_forecast_observation
    elif request.param == 'probfxobsy':
        return single_prob_forecast_observation_y
    elif request.param == 'probfxagg':
        return single_prob_forecast_aggregate
    elif request.param == 'probfx_ref':
        return single_prob_forecast_observation_reffx
    elif request.param == 'probfxagg_ref':
        return single_prob_forecast_aggregate_reffx


@pytest.fixture()
def proc_fx_obs(create_processed_fxobs, many_forecast_observation):
    proc_fx_obs = []

    for fx_obs in many_forecast_observation:
        proc_fx_obs.append(
            create_processed_fxobs(fx_obs,
                                   np.random.randn(10)+10,
                                   np.random.randn(10)+10)
        )
    return proc_fx_obs


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.parametrize('categories,metrics', [
    pytest.param(
        [], [], marks=pytest.mark.xfail(strict=True, type=RuntimeError)),
    pytest.param(
        ['date', 'month'], [],
        marks=pytest.mark.xfail(strict=True, type=RuntimeError)),
    (LIST_OF_CATEGORIES, DETERMINISTIC_METRICS),
    (LIST_OF_CATEGORIES, DET_NO_REF),
])
def test_calculate_metrics_with_reference(
        categories, metrics, proc_fx_obs):
    result = calculator.calculate_metrics(proc_fx_obs,
                                          categories, metrics)

    assert isinstance(result, list)
    assert isinstance(result[0], datamodel.MetricResult)
    assert len(result) == len(proc_fx_obs)


def _random_ref_values_if_not_None(data_obj):
    if data_obj.reference_forecast is not None:
        return np.random.randn(10) + 10
    else:
        return None


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_calculate_metrics_single(single_forecast_data_obj,
                                  create_processed_fxobs):
    ref_values = _random_ref_values_if_not_None(single_forecast_data_obj)
    inp = [create_processed_fxobs(single_forecast_data_obj,
                                  np.random.randn(10)+10,
                                  np.random.randn(10)+10,
                                  ref_values=ref_values)]
    result = calculator.calculate_metrics(inp, LIST_OF_CATEGORIES,
                                          DETERMINISTIC_METRICS)
    assert isinstance(result, list)
    assert isinstance(result[0], datamodel.MetricResult)
    assert len(result) == 1


def test_calculate_metrics_single_uncert(single_forecast_observation_uncert,
                                         create_processed_fxobs):
    inp = [create_processed_fxobs(single_forecast_observation_uncert,
                                  np.array([1.9, 1, 1.0005]),
                                  np.array([1, 1, 1]))]
    result = calculator.calculate_metrics(inp, ['total'], ['mae'])
    assert isinstance(result, list)
    assert isinstance(result[0], datamodel.MetricResult)
    assert len(result) == 1
    expected_values = {None: (0.9 + 0.0005)/3, 100.: 0., 0.1: 0.3}
    expected = expected_values[single_forecast_observation_uncert.uncertainty]
    assert result[0].values[0].value == expected


def test_calculate_metrics_explicit_cost(single_forecast_observation_uncert,
                                         create_processed_fxobs,
                                         constant_cost):
    inp = [create_processed_fxobs(single_forecast_observation_uncert,
                                  np.array([1.9, 1, 1.0005]),
                                  np.array([1, 1, 1]))]
    result = calculator.calculate_metrics(inp, ['total'], ['cost'])
    assert isinstance(result, list)
    assert isinstance(result[0], datamodel.MetricResult)
    assert len(result) == 1
    expected_values = {None: (0.9 + 5e-4), 100.: 0., 0.1: 0.3 * 3}
    expected = expected_values[single_forecast_observation_uncert.uncertainty]
    # float precision issues
    assert abs(result[0].values[0].value - expected) < 1e-8


def test_calculate_deterministic_metrics_sorting(single_forecast_observation,
                                                 create_processed_fxobs,
                                                 create_dt_index):
    index = pd.DatetimeIndex(
        # sunday, monday, tuesday
        ['20200531 1900Z', '20200601 2000Z', '20200602 2100Z'])
    inp = create_processed_fxobs(
        single_forecast_observation,
        pd.Series([2, 1, 0], index=index),
        pd.Series([1, 1, 1], index=index))
    categories = ('hour', 'total', 'date', 'month', 'weekday')
    metrics = ('rmse', 'ksi', 'mbe', 'mae')
    result = calculator.calculate_deterministic_metrics(
        inp, categories, metrics)
    expected = {
        0: ('total', 'mae', '0', 2/3),
        1: ('total', 'mbe', '0', 0.),
        4: ('month', 'mae', 'May', 1.),
        5: ('month', 'mae', 'Jun', 0.5),
        12: ('hour', 'mae', '19', 1.),
        14: ('hour', 'mae', '21', 1.),
        18: ('hour', 'rmse', '19', 1.),
        39: ('weekday', 'mbe', 'Mon', 0.),
        40: ('weekday', 'mbe', 'Tue', -1.),
        41: ('weekday', 'mbe', 'Sun', 1.),
    }
    attr_order = ('category', 'metric', 'index', 'value')
    for k, expected_attrs in expected.items():
        for attr, expected_val in zip(attr_order, expected_attrs):
            assert getattr(result.values[k], attr) == expected_val


def test_calculate_event_metrics_sorting(single_event_forecast_observation,
                                         create_processed_fxobs,
                                         create_dt_index):
    inp = create_processed_fxobs(
        single_event_forecast_observation,
        pd.Series([True, False, True], index=create_dt_index(3)),  # fx
        pd.Series([True, True, True], index=create_dt_index(3)))   # obs
    categories = ('hour', 'total')
    metrics = ('far', 'pod', 'ea', 'pofd')
    result = calculator.calculate_event_metrics(
        inp, categories, metrics)
    expected = {
        0: ('total', 'pod', '0', 2 / 3),
        1: ('total', 'far', '0', 0.0),
        2: ('total', 'pofd', '0', 0.0),
        3: ('total', 'ea', '0', 2 / 3),
        4: ('hour', 'pod', '0', 1.0),
        6: ('hour', 'pod', '2', 1.0),
        8: ('hour', 'far', '1', 0.0),
        10: ('hour', 'pofd', '0', 0.0),
        15: ('hour', 'ea', '2', 1.0),
    }
    attr_order = ('category', 'metric', 'index', 'value')
    for k, expected_attrs in expected.items():
        for attr, expected_val in zip(attr_order, expected_attrs):
            assert getattr(result.values[k], attr) == expected_val


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_calculate_metrics_with_probablistic(single_observation,
                                             prob_forecasts,
                                             create_processed_fxobs,
                                             create_dt_index,
                                             copy_prob_forecast_with_axis,
                                             caplog):
    const_values = [10, 20, 30]
    conv_prob_fx = copy_prob_forecast_with_axis(
        prob_forecasts, prob_forecasts.axis, constant_values=const_values)
    prfxobs = datamodel.ForecastObservation(
        conv_prob_fx, single_observation)
    fx_values = pd.DataFrame(np.array((
        np.linspace(0, 9., 10),
        np.linspace(1, 10., 10),
        np.linspace(2, 11., 10))).T,
        columns=const_values,
        index=create_dt_index(10))
    obs_values = pd.Series(
        np.linspace(1.5, 10.5, 10),
        index=create_dt_index(10))
    proc_prfx_obs = create_processed_fxobs(prfxobs, fx_values, obs_values)

    # Without reference
    result = calculator.calculate_metrics([proc_prfx_obs],
                                          LIST_OF_CATEGORIES,
                                          PROB_NO_REF)
    assert len(result) == 1
    assert isinstance(result[0], datamodel.MetricResult)
    verify_metric_result(result[0],
                         proc_prfx_obs,
                         LIST_OF_CATEGORIES,
                         probabilistic._REQ_DIST)

    # With reference
    ref_fx_values = fx_values + .5
    conv_ref_prob_fx = copy_prob_forecast_with_axis(
        prob_forecasts, prob_forecasts.axis, constant_values=const_values)
    ref_prfxobs = datamodel.ForecastObservation(
        conv_ref_prob_fx,
        single_observation,
        reference_forecast=conv_ref_prob_fx)
    proc_ref_prfx_obs = create_processed_fxobs(ref_prfxobs,
                                               fx_values,
                                               obs_values,
                                               ref_values=ref_fx_values)
    dist_results = calculator.calculate_metrics(
        [proc_ref_prfx_obs], LIST_OF_CATEGORIES,
        # reverse to ensure order output is independent
        PROBABILISTIC_METRICS[::-1])

    assert 'Failed' not in caplog.text
    assert isinstance(dist_results, list)
    assert len(dist_results) == 1
    assert isinstance(dist_results[0], datamodel.MetricResult)
    verify_metric_result(dist_results[0],
                         proc_ref_prfx_obs,
                         LIST_OF_CATEGORIES,
                         probabilistic._REQ_DIST)

    expected = {
        0: ('total', 'crps', '0', 17.247),
        1: ('year', 'crps', '2019', 17.247),
        2: ('month', 'crps', 'Aug', 17.247),
        3: ('hour', 'crps', '0', 19.801000000000002),
        4: ('hour', 'crps', '1', 19.405)
    }
    attr_order = ('category', 'metric', 'index', 'value')
    for k, expected_attrs in expected.items():
        for attr, expected_val in zip(attr_order, expected_attrs):
            assert getattr(dist_results[0].values[k], attr) == expected_val

    # Constant Values
    cv_proc_ref_prfx_obs = []
    zip_cvs = zip(conv_prob_fx.constant_values,
                  conv_ref_prob_fx.constant_values)
    for i, (cv, ref_cv) in enumerate(zip_cvs):
        ref_fx_values = fx_values.iloc[:, i] + .5
        cv_ref_prfxobs = datamodel.ForecastObservation(
            cv,
            single_observation,
            reference_forecast=ref_cv)
        cv_proc_ref_prfx_obs.append(
            create_processed_fxobs(cv_ref_prfxobs,
                                   fx_values.iloc[:, i],
                                   obs_values,
                                   ref_values=ref_fx_values))

    cv_results = calculator.calculate_metrics(
        cv_proc_ref_prfx_obs, LIST_OF_CATEGORIES,
        # reverse to ensure order output is independent
        list(PROB_NO_DIST)[::-1])

    # test MetricValues contents and order
    single_result = cv_results[0]
    expected = {
        0: ('total', 'bs', '0', 0.8308500000000001),
        1: ('total', 'bss', '0', -0.010366947374821578),
        7: ('year', 'bs', '2019', 0.8308500000000001),
        11: ('year', 'unc', '2019', 0.08999999999999998),
        14: ('month', 'bs', 'Aug', 0.8308500000000001)
    }
    attr_order = ('category', 'metric', 'index', 'value')
    for k, expected_attrs in expected.items():
        for attr, expected_val in zip(attr_order, expected_attrs):
            assert getattr(single_result.values[k], attr) == expected_val

    # Distribution and constant values
    all_results = calculator.calculate_metrics(
        [proc_prfx_obs] + cv_proc_ref_prfx_obs, LIST_OF_CATEGORIES,
        # reverse to ensure order output is independent
        PROBABILISTIC_METRICS[::-1])
    assert all_results[0] == dist_results[0]
    assert len(all_results[1:]) == len(cv_results)
    for a, b in zip(all_results[1:], cv_results):
        assert a == b


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_calculate_metrics_with_probablistic_dist_missing_ref(
        single_observation, prob_forecasts, create_processed_fxobs,
        create_dt_index, copy_prob_forecast_with_axis, caplog):
    const_values = [10, 20, 30]
    conv_prob_fx = copy_prob_forecast_with_axis(
        prob_forecasts, prob_forecasts.axis, constant_values=const_values)
    prfxobs = datamodel.ForecastObservation(conv_prob_fx, single_observation)

    fx_values = pd.DataFrame(np.random.randn(10, 3)+10,
                             index=create_dt_index(10))
    obs_values = pd.Series(np.random.randn(10)+10,
                           index=create_dt_index(10))
    pairs = []
    for i, cv in enumerate(prfxobs.forecast.constant_values):
        cv_fxobs = datamodel.ForecastObservation(
            cv, single_observation)
        pairs.append(create_processed_fxobs(
            cv_fxobs, fx_values[i], obs_values))
    results = calculator.calculate_metrics(
        pairs, LIST_OF_CATEGORIES, probabilistic._REQ_REF_FX)
    assert len(results) == 3
    for res in results:
        for val in res.values:
            assert np.isnan(val.value)


def test_calculate_deterministic_metrics_no_metrics(
        create_processed_fxobs, single_forecast_observation):
    proc_fx_obs = create_processed_fxobs(single_forecast_observation,
                                         np.array([1]), np.array([1]))
    with pytest.raises(RuntimeError):
        calculator.calculate_deterministic_metrics(
            proc_fx_obs, LIST_OF_CATEGORIES, []
        )


def test_calculate_deterministic_metrics_no_cost_param(
        create_processed_fxobs, single_forecast_observation):
    proc_fx_obs = create_processed_fxobs(single_forecast_observation,
                                         np.array([1]), np.array([1])).replace(
                                             cost=None)

    out = calculator.calculate_deterministic_metrics(
            proc_fx_obs, LIST_OF_CATEGORIES, ['cost']
    )
    for o in out.values:
        assert np.isnan(o.value)


@pytest.mark.parametrize('fx_vals,obs_vals', [
    (np.random.randn(0), np.random.randn(0)),
    (np.random.randn(10), np.random.randn(0)),
    (np.random.randn(0), np.random.randn(10)),
])
def test_calculate_deterministic_metrics_missing_values(
        create_processed_fxobs, single_forecast_observation,
        fx_vals, obs_vals):
    pair = create_processed_fxobs(single_forecast_observation,
                                  fx_vals, obs_vals)
    with pytest.raises(RuntimeError):
        calculator.calculate_deterministic_metrics(
            pair, LIST_OF_CATEGORIES, DET_NO_REF
        )


def test_calculate_deterministic_metrics_normalizer(
        create_processed_fxobs, single_forecast_observation_norm):
    pair = create_processed_fxobs(single_forecast_observation_norm,
                                  np.random.randn(10), np.random.randn(10))
    s_normed = calculator.calculate_deterministic_metrics(
        pair, ['total'], deterministic._REQ_NORM)
    unnormed = [x.lstrip('n') for x in deterministic._REQ_NORM]
    s_unnormed = calculator.calculate_deterministic_metrics(
        pair, ['total'], unnormed)
    norm = single_forecast_observation_norm.normalization
    for v_normed, v_unnormed in zip(s_normed.values, s_unnormed.values):
        if np.isnan(norm):
            assert np.isnan(v_normed.value)
        else:
            assert_allclose(v_normed.value, v_unnormed.value * 100 / norm)


def test_calculate_deterministic_metrics_reference(
        create_processed_fxobs, single_forecast_observation_reffx):
    pair0 = create_processed_fxobs(single_forecast_observation_reffx,
                                   np.random.randn(10), np.random.randn(10),
                                   ref_values=np.random.randn(10))
    pair1 = create_processed_fxobs(single_forecast_observation_reffx,
                                   np.random.randn(10), np.random.randn(10),
                                   ref_values=np.random.randn(10))
    s0 = calculator.calculate_deterministic_metrics(
        pair0, ['total'], deterministic._REQ_REF_FX)
    s1 = calculator.calculate_deterministic_metrics(
        pair1, ['total'], deterministic._REQ_REF_FX)
    for s in [s0, s1]:
        assert isinstance(s, datamodel.MetricResult)
    assert s0 != s1


def verify_metric_result(result, pair, categories, metrics):
    assert result.name == pair.original.forecast.name
    if not categories or not metrics:
        return
    cats = {val.category for val in result.values}
    assert cats == set(categories)
    fx_values = pair.forecast_values
    for cat in categories:
        cat_grps = {v.index for v in result.values if v.category == cat}
        assert len(
            {v.metric for v in result.values if v.category == cat}
        ) == len(metrics)

        # has expected groupings
        if cat == 'month':
            grps = fx_values.groupby(
                fx_values.index.month).groups
            grps = [calendar.month_abbr[g] for g in grps]
        elif cat == 'hour':
            grps = fx_values.groupby(
                fx_values.index.hour).groups
        elif cat == 'year':
            grps = fx_values.groupby(
                fx_values.index.year).groups
        elif cat == 'date':
            grps = fx_values.groupby(
                fx_values.index.date).groups
        elif cat == 'weekday':
            grps = fx_values.groupby(
                fx_values.index.weekday).groups
            grps = [calendar.day_abbr[g] for g in grps]
        elif cat == 'total':
            grps = ['0']
        assert {str(g) for g in grps} == cat_grps
    for val in result.values:
        assert (
            np.isnan(val.value) or
            np.issubdtype(type(val.value), np.number)
        )


# Suppress RuntimeWarnings b/c in some metrics will divide by zero or
# don't handle single values well
@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.parametrize('categories', [
    [],
    *list(itertools.combinations(LIST_OF_CATEGORIES, 1)),
    LIST_OF_CATEGORIES[0:1],
    LIST_OF_CATEGORIES[0:2],
    LIST_OF_CATEGORIES
])
@pytest.mark.parametrize('metrics', [
    *list(itertools.combinations(DETERMINISTIC_METRICS, 1)),
    DETERMINISTIC_METRICS[0:1],
    DETERMINISTIC_METRICS[0:2],
    DETERMINISTIC_METRICS
])
def test_calculate_deterministic_metrics(categories, metrics,
                                         single_forecast_data_obj,
                                         single_forecast_observation,
                                         create_processed_fxobs):
    ref_values = _random_ref_values_if_not_None(single_forecast_data_obj)
    pair = create_processed_fxobs(single_forecast_data_obj,
                                  np.random.randn(10)+10,
                                  np.random.randn(10)+10,
                                  ref_values=ref_values)
    result = calculator.calculate_deterministic_metrics(
        pair, categories, metrics)
    # Check results
    assert isinstance(result, datamodel.MetricResult)
    verify_metric_result(result, pair, categories, metrics)


def test_calculate_probabilistic_metrics_no_metrics(
        single_prob_forecast_observation, create_processed_fxobs):
    proc_fxobs = create_processed_fxobs(single_prob_forecast_observation,
                                        pd.DataFrame(np.random.randn(10, 3)),
                                        pd.Series(np.random.randn(10)))
    with pytest.raises(RuntimeError):
        calculator.calculate_probabilistic_metrics(
            proc_fxobs, LIST_OF_CATEGORIES, []
        )


# numpy < operator warns on comparison with nan
@pytest.mark.filterwarnings("ignore:invalid value")
def test_calculate_probabilistic_metrics_missing_ref(
        single_prob_forecast_observation, create_processed_fxobs,
        create_dt_index):
    proc_fxobs = create_processed_fxobs(
        single_prob_forecast_observation,
        pd.DataFrame(np.random.randn(10, 3), index=create_dt_index(10)),
        pd.Series(np.random.randn(10), index=create_dt_index(10)))
    result = calculator.calculate_probabilistic_metrics(
            proc_fxobs, LIST_OF_CATEGORIES, probabilistic._REQ_REF_FX
        )
    assert isinstance(result, datamodel.MetricResult)
    assert result.values == ()


# numpy < operator warns on comparison with nan
@pytest.mark.filterwarnings("ignore:invalid value")
def test_calculate_probabilistic_metrics_no_reference_data(
        single_prob_forecast_observation_reffx, create_processed_fxobs,
        create_dt_index):
    proc_fxobs = create_processed_fxobs(
        single_prob_forecast_observation_reffx,
        pd.DataFrame(np.random.randn(10, 3), index=create_dt_index(10)),
        pd.Series(np.random.randn(10), index=create_dt_index(10)),
        ref_values=None)
    result = calculator.calculate_probabilistic_metrics(
            proc_fxobs, LIST_OF_CATEGORIES, probabilistic._REQ_REF_FX
        )
    assert isinstance(result, datamodel.MetricResult)
    assert result.values == ()


def test_calculate_probabilistic_metrics_interval_label_ending(
        single_prob_forecast_observation, create_processed_fxobs,
        create_dt_index):
    proc_fxobs = create_processed_fxobs(
        single_prob_forecast_observation,
        pd.DataFrame(np.random.randn(10, 3), index=create_dt_index(10)),
        pd.Series(np.random.randn(10), index=create_dt_index(10))
    )
    proc_fxobs = proc_fxobs.replace(interval_label='ending')
    result = calculator.calculate_probabilistic_metrics(
        proc_fxobs, LIST_OF_CATEGORIES, PROB_NO_REF)
    assert result


def test_calculate_probabilistic_metrics_bad_reference_axis(
        single_prob_forecast_observation, prob_forecasts, single_observation,
        create_processed_fxobs,
        copy_prob_forecast_with_axis):
    conv_fx = copy_prob_forecast_with_axis(prob_forecasts, axis='y')
    fxob = datamodel.ForecastObservation(
        prob_forecasts,
        single_observation,
        conv_fx)
    proc_fxobs = create_processed_fxobs(
        fxob,
        pd.DataFrame(np.random.randn(10, 3)),
        pd.Series(np.random.randn(10)),
        ref_values=pd.DataFrame(np.random.randn(10, 3)),)
    with pytest.raises(ValueError):
        calculator.calculate_probabilistic_metrics(
            proc_fxobs, LIST_OF_CATEGORIES, probabilistic._REQ_REF_FX
        )


def test_calculate_probabilistic_metrics_missing_values(
        single_prob_forecast_observation, create_processed_fxobs):
    proc_fxobs = create_processed_fxobs(single_prob_forecast_observation,
                                        pd.DataFrame(),
                                        pd.Series(np.random.randn(10)))
    with pytest.raises(RuntimeError):
        calculator.calculate_probabilistic_metrics(
            proc_fxobs, LIST_OF_CATEGORIES, PROB_NO_REF
        )


def test_calculate_probabilistic_metrics_missing_observation(
        single_prob_forecast_observation, create_processed_fxobs):
    proc_fxobs = create_processed_fxobs(single_prob_forecast_observation,
                                        pd.DataFrame(np.random.randn(10, 3)),
                                        pd.Series([], dtype=float))
    with pytest.raises(RuntimeError):
        calculator.calculate_probabilistic_metrics(
            proc_fxobs, LIST_OF_CATEGORIES, PROB_NO_REF
        )


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_calculate_probabilistic_metrics_dist_simple(prob_forecasts_data_obj,
                                                     create_processed_fxobs,
                                                     create_dt_index):
    if prob_forecasts_data_obj.reference_forecast is None:
        ref_values = None
    else:
        ref_values = pd.DataFrame(
            np.random.randn(10, 3)+10, index=create_dt_index(10))
    pair = create_processed_fxobs(
        prob_forecasts_data_obj,
        pd.DataFrame(np.random.randn(10, 3)+10, index=create_dt_index(10)),
        pd.Series(np.random.randn(10)+10, index=create_dt_index(10)),
        ref_values=ref_values)
    result = calculator.calculate_probabilistic_metrics(
        pair, LIST_OF_CATEGORIES, PROBABILISTIC_METRICS)
    assert isinstance(result, datamodel.MetricResult)
    verify_metric_result(
        result, pair, LIST_OF_CATEGORIES, probabilistic._REQ_DIST)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_calculate_probabilistic_metrics_with_constant_value_simple(
        prob_forecasts_data_obj,
        create_processed_fxobs,
        create_dt_index):
    if prob_forecasts_data_obj.reference_forecast is None:
        ref_values = None
    else:
        ref_values = pd.DataFrame(
            np.random.randn(10, 3)+10, index=create_dt_index(10))
    pair = create_processed_fxobs(
        prob_forecasts_data_obj,
        pd.DataFrame(np.random.randn(10, 3)+10, index=create_dt_index(10)),
        pd.Series(np.random.randn(10)+10, index=create_dt_index(10)),
        ref_values=ref_values)

    for i, cv in enumerate(pair.original.forecast.constant_values):
        data_object = prob_forecasts_data_obj.data_object.replace(
            **{'interval_length': cv.interval_length})
        if isinstance(data_object, datamodel.Aggregate):
            pair_type = datamodel.ForecastAggregate
        else:
            pair_type = datamodel.ForecastObservation
        fxobs = pair_type(
            cv,
            data_object,
            pair.original.reference_forecast)
        fx = pair.forecast_values[i]
        fx.name = 'value'
        cv_pair = create_processed_fxobs(
            fxobs,
            fx,
            pair.observation_values,
            ref_values=ref_values)
        result = calculator.calculate_probabilistic_metrics(
            cv_pair, LIST_OF_CATEGORIES, PROBABILISTIC_METRICS)
        assert isinstance(result, datamodel.MetricResult)
        verify_metric_result(result, cv_pair, LIST_OF_CATEGORIES, PROB_NO_DIST)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.parametrize('categories', [
    [],
    *list(itertools.combinations(LIST_OF_CATEGORIES, 1)),
    LIST_OF_CATEGORIES[0:1],
    LIST_OF_CATEGORIES[0:2],
    LIST_OF_CATEGORIES
])
@pytest.mark.parametrize('metrics', [
    *list(itertools.combinations(PROBABILISTIC_METRICS, 1)),
    PROBABILISTIC_METRICS[0:1],
    PROBABILISTIC_METRICS[0:2],
    PROBABILISTIC_METRICS
])
@pytest.mark.parametrize('axis,prob_fx_df,ref_fx_df,obs', [
    ('y',
     pd.DataFrame({'25': np.random.randn(10)*np.sqrt(20)+20,
                   '50': np.random.randn(10)*np.sqrt(20)+20,
                   '75': np.random.randn(10)*np.sqrt(20)+20}, dtype=float),
     pd.DataFrame({'25': np.random.randn(10)*np.sqrt(20)+20,
                   '50': np.random.randn(10)*np.sqrt(20)+20,
                   '75': np.random.randn(10)*np.sqrt(20)+20}, dtype=float),
     pd.Series(np.random.randn(10)*np.sqrt(20)+20, dtype=float)),
    ('x',
     pd.DataFrame({'10': np.random.randn(10)*5+25,
                   '20': np.random.randn(10)*5+50,
                   '30': np.random.randn(10)*5+75}, dtype=float),
     pd.DataFrame({'10': np.random.randn(10)*5+25,
                   '20': np.random.randn(10)*5+50,
                   '30': np.random.randn(10)*5+75}, dtype=float),
     pd.Series(np.random.randn(10)*np.sqrt(20)+20, dtype=float))
])
def test_calculate_probabilistic_metrics(categories, metrics,
                                         axis, prob_fx_df, ref_fx_df, obs,
                                         prob_forecasts, single_observation,
                                         copy_prob_forecast_with_axis,
                                         create_processed_fxobs,
                                         create_dt_index):
    # add index to data
    dt_index = create_dt_index(len(prob_fx_df))
    prob_fx_df.index = dt_index
    ref_fx_df.index = dt_index
    obs.index = dt_index

    conv_prob_fx = copy_prob_forecast_with_axis(
        prob_forecasts, axis, prob_fx_df.columns.tolist())
    conv_ref_fx = copy_prob_forecast_with_axis(
        prob_forecasts, axis, prob_fx_df.columns.tolist())

    # Full distributions
    if any(x for x in PROB_NO_DIST if x in set(metrics)):
        # create processed pairs
        prob_fxobs = datamodel.ForecastObservation(
            conv_prob_fx, single_observation, reference_forecast=conv_ref_fx)
        pair = create_processed_fxobs(
            prob_fxobs, prob_fx_df, obs, ref_values=ref_fx_df)

        # calculation
        dist_result = calculator.calculate_probabilistic_metrics(
            pair, categories, metrics)
        metrics_sans_dist = list(set(metrics) - set(PROB_NO_DIST))
        assert isinstance(dist_result, datamodel.MetricResult)
        verify_metric_result(dist_result, pair, categories, metrics_sans_dist)

    # Each constant value
    if any(x for x in PROB_NO_DIST if x in set(metrics)):
        for i, cv in enumerate(conv_prob_fx.constant_values):
            # create processed pairs
            cv_series = prob_fx_df[str(cv.constant_value)]
            cv_ref_series = ref_fx_df[str(cv.constant_value)]
            cv_prob_fxobs = datamodel.ForecastObservation(
                cv, single_observation, reference_forecast=cv)
            cv_pair = create_processed_fxobs(
                cv_prob_fxobs, cv_series, obs, ref_values=cv_ref_series)

            # calculation
            cv_result = calculator.calculate_probabilistic_metrics(
                cv_pair, categories, metrics)
            assert isinstance(cv_result, datamodel.MetricResult)
            cv_metrics = list(set(metrics) - set(probabilistic._REQ_DIST))
            verify_metric_result(cv_result, cv_pair, categories, cv_metrics)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.parametrize('categories', [
    [],
    *list(itertools.combinations(LIST_OF_CATEGORIES, 1)),
    LIST_OF_CATEGORIES[0:1],
    LIST_OF_CATEGORIES[0:2],
    LIST_OF_CATEGORIES
])
@pytest.mark.parametrize('interval_label', ['beginning', 'ending'])
@pytest.mark.parametrize('df,metrics', [
    (pd.DataFrame({
        ('50', 'forecast'): np.random.randn(10),
        ('50', 'probability'): np.random.randn(10),
        ('50', 'reference_forecast'): np.random.randn(10),
        ('50', 'reference_probability'): np.random.randn(10),
        (None, 'observation'): np.random.randn(10)
        }),
     PROB_NO_DIST),
    (pd.DataFrame({
        ('50', 'forecast'): np.random.randn(10),
        ('50', 'probability'): np.random.randn(10),
        (None, 'observation'): np.random.randn(10)
        }),
     PROB_NO_REF_NO_DIST),
    (pd.DataFrame({
        ('25', 'forecast'): np.random.randn(10),
        ('50', 'forecast'): np.random.randn(10),
        ('75', 'forecast'): np.random.randn(10),
        ('25', 'probability'): np.random.randn(10),
        ('50', 'probability'): np.random.randn(10),
        ('75', 'probability'): np.random.randn(10),
        ('25', 'reference_forecast'): np.random.randn(10),
        ('50', 'reference_forecast'): np.random.randn(10),
        ('75', 'reference_forecast'): np.random.randn(10),
        ('25', 'reference_probability'): np.random.randn(10),
        ('50', 'reference_probability'): np.random.randn(10),
        ('75', 'reference_probability'): np.random.randn(10),
        (None, 'observation'): np.random.randn(10)
        }),
     probabilistic._REQ_DIST),
    (pd.DataFrame({
        ('25', 'forecast'): np.random.randn(10),
        ('50', 'forecast'): np.random.randn(10),
        ('75', 'forecast'): np.random.randn(10),
        ('25', 'probability'): np.random.randn(10),
        ('50', 'probability'): np.random.randn(10),
        ('75', 'probability'): np.random.randn(10),
        (None, 'observation'): np.random.randn(10)
        }),
     probabilistic._REQ_DIST)
])
def test_calculate_probabilistic_metrics_from_df(categories, df, metrics,
                                                 interval_label,
                                                 create_dt_index):
    df.index = create_dt_index(len(df))
    results = calculator._calculate_probabilistic_metrics_from_df(
        df, categories, metrics, interval_label)
    assert all(isinstance(r, datamodel.MetricValue) for r in results)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.parametrize('metric,fx,obs,ref_fx,norm,dead,expect', [
    ('mae', [], [], None, None, None, np.NaN),
    ('mae', [], [], None, None, 1, np.NaN),
    ('mae', [1, 1, 1], [0, 1, -1], None, None, None, 1.0),
    ('mae', [0, 1, 4], [1, 1, 1], None, None, 100, 1.),
    ('mbe', [1, 1, 1], [0, 1, -1], None, None, None, 1.0),
    ('rmse', [1, 0, 1], [0, -1, 2], None, None, None, 1.0),
    ('nrmse', [1, 0, 1], [0, -1, 2], None, None, None, None),
    ('nrmse', [1, 0, 1], [0, -1, 2], None, 2.0, None, 100/2),
    ('nrmse', [0.9, 1, 0.1], [1, 0, 1], None, 2.0, 100, 100*np.sqrt(1/3)/2),
    ('mape', [2, 3, 1], [4, 2, 2], None, None, None, 50.0),
    ('s', [1, 0, 1], [0, -1, 2], None, None, None, None),
    ('s', [1, 0, 1], [0, -1, 2], [2, 1, 0], None, None, 0.5),
    ('s', [1, 0, 1], [0, -1, 2], [2, 1, 0], None, 100, 0.6464466094067263),
    ('r', [3, 2, 1], [1, 2, 3], None, None, None, -1.0),
    ('r^2', [3, 2, 1], [1, 2, 3], None, None, None, -3.0),
    ('crmse', [1, 1, 1], [0, 1, 2], None, None, None, np.sqrt(2/3))
])
def test_apply_deterministic_metric_func(metric, fx, obs, ref_fx, norm,
                                         dead, expect):
    fx_series = np.array(fx)
    obs_series = np.array(obs)
    # Check require reference forecast kwarg
    if metric in ['s']:
        if ref_fx is None:
            with pytest.raises(KeyError):
                # Missing positional argument
                calculator._apply_deterministic_metric_func(
                    metric, fx_series, obs_series)
        else:
            ref_fx_series = np.array(ref_fx)
            metric_value = calculator._apply_deterministic_metric_func(
                metric, fx_series, obs_series, ref_fx=ref_fx_series,
                deadband=dead)
            np.testing.assert_approx_equal(metric_value, expect)

    # Check requires normalization kwarg
    elif metric in ['nrmse']:
        if norm is None:
            with pytest.raises(KeyError):
                # Missing positional argument
                calculator._apply_deterministic_metric_func(
                    metric, fx_series, obs_series)
        else:
            metric_value = calculator._apply_deterministic_metric_func(
                metric, fx_series, obs_series, normalization=norm,
                deadband=dead)
            np.testing.assert_approx_equal(metric_value, expect)

    # Does not require kwarg
    else:
        metric_value = calculator._apply_deterministic_metric_func(
            metric, fx_series, obs_series, deadband=dead)
        if np.isnan(expect):
            assert np.isnan(metric_value)
        else:
            np.testing.assert_approx_equal(metric_value, expect)


def test_apply_deterministic_bad_metric_func():
    with pytest.raises(KeyError):
        calculator._apply_deterministic_metric_func('BAD METRIC',
                                                    np.array([]),
                                                    np.array([]))


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.parametrize('metric,fx,fx_prob,obs,ref_fx,ref_fx_prob,expect', [
    ('bs', [], [], [], None, None, np.NaN),
    ('bs', [1, 1, 1], [100, 100, 100], [1, 1, 1], None, None, 0.),

    # Briar Skill Score with no reference
    ('bss', [1, 1, 1], [100, 100, 100], [0, 0, 0],
        None, None, 1.),
    ('bss', [1, 1, 1], [100, 100, 100], [1, 1, 1],
        [0, 0, 0], [100, 100, 100], 1.),

    ('rel', [1, 1, 1], [100, 100, 100], [1, 1, 1], None, None, 0.),
    ('res', [1, 1, 1], [100, 100, 100], [1, 1, 1], None, None, 0.),
    ('unc', [1, 1, 1], [100, 100, 100], [1, 1, 1], None, None, 0.),

    # CRPS single forecast
    ('crps', [[1]], [[100]], [[0]], None, None, 0.),
    # CRPS mulitple forecasts
    ('crps', [[1, 1, 1], [2, 2, 2], [3, 3, 3]],
             [[100, 100, 100], [100, 100, 100], [100, 100, 100]],
             [0, 0, 0], None, None, 0.)
])
def test_apply_probabilistic_metric_func(metric, fx, fx_prob, obs,
                                         ref_fx, ref_fx_prob, expect,
                                         create_dt_index):
    fx_data = np.array(fx)
    fx_prob_data = np.array(fx_prob)
    obs_series = np.array(obs)

    # Check metrics that require reference forecast kwarg
    if metric in probabilistic._REQ_REF_FX:
        if ref_fx is None or ref_fx_prob is None:
            with pytest.raises(KeyError):
                # Missing positional argument
                calculator._apply_probabilistic_metric_func(
                    metric, fx_data, fx_prob_data, obs_series)
        else:
            ref_fx_data = np.array(ref_fx)
            ref_fx_prob_data = np.array(ref_fx_prob)
            metric_value = calculator._apply_probabilistic_metric_func(
                metric, fx_data, fx_prob_data, obs_series,
                ref_fx=ref_fx_data,
                ref_fx_prob=ref_fx_prob_data)
            np.testing.assert_approx_equal(metric_value, expect)

    # Does not require kwarg
    else:
        metric_value = calculator._apply_probabilistic_metric_func(
            metric, fx_data, fx_prob_data, obs_series)
        if np.isnan(expect):
            assert np.isnan(metric_value)
        else:
            np.testing.assert_approx_equal(metric_value, expect)


def test_apply_probabilistic_bad_metric_func():
    with pytest.raises(KeyError):
        calculator._apply_probabilistic_metric_func('BAD METRIC',
                                                    np.array([]),
                                                    np.array([]),
                                                    np.array([]))


@pytest.mark.parametrize('ts,tz,interval_label,category,result', [
    # category: hour
    (
        ['20191210T1300', '20191210T1330', '20191210T1400'],
        "UTC",
        'ending',
        'hour',
        {('12', -1.0), ('13', 0.5)}
    ),
    (
        ['20191210T1300', '20191210T1330', '20191210T1400'],
        "UTC",
        'beginning',
        'hour',
        {('13', -0.5), ('14', 1.0)}
    ),
    (
        ['20191210T1300', '20191210T1330', '20191210T1400'],
        "US/Pacific",
        'ending',
        'hour',
        {('4', -1.0), ('5', 0.5)}
    ),
    (
        ['20191210T1300', '20191210T1330', '20191210T1400'],
        "US/Pacific",
        'beginning',
        'hour',
        {('5', -0.5), ('6', 1.0)}
    ),

    # category: month
    (
        ['20191130T2330', '20191201T0000', '20191201T0030'],
        "UTC",
        'ending',
        'month',
        {('Nov', -0.5), ('Dec', 1.0)}
    ),
    (
        ['20191130T2330', '20191201T0000', '20191201T0030'],
        "UTC",
        'beginning',
        'month',
        {('Nov', -1.0), ('Dec', 0.5)}
    ),

    # category: year
    (
        ['20191231T2330', '20200101T0000', '20200101T0030'],
        "UTC",
        'ending',
        'year',
        {('2019', -0.5), ('2020', 1.0)}
    ),
    (
        ['20191231T2330', '20200101T0000', '20200101T0030'],
        "UTC",
        'beginning',
        'year',
        {('2019', -1.0), ('2020', 0.5)}
    ),
])
def test_interval_label(ts, tz, interval_label, category, result,
                        create_processed_fxobs):

    index = pd.DatetimeIndex(ts, tz="UTC")

    # Custom metadata to keep all timestamps in UTC for tests
    site = datamodel.Site(
        name='Albuquerque Baseline',
        latitude=35.05,
        longitude=-106.54,
        elevation=1657.0,
        timezone="UTC",
        provider='Sandia'
    )

    fx_series = pd.Series([0, 1, 2], index=index)
    obs_series = pd.Series([1, 1, 1], index=index)

    # convert to local timezone
    fx_series = fx_series.tz_convert(tz)
    obs_series = obs_series.tz_convert(tz)

    fxobs = datamodel.ForecastObservation(
        observation=datamodel.Observation(
            site=site, name='dummy obs', variable='ghi',
            interval_value_type='instantaneous', uncertainty=1,
            interval_length=pd.Timedelta(obs_series.index.freq),
            interval_label=interval_label
        ),
        forecast=datamodel.Forecast(
            site=site, name='dummy fx', variable='ghi',
            interval_value_type='instantaneous',
            interval_length=pd.Timedelta(fx_series.index.freq),
            interval_label=interval_label,
            issue_time_of_day=datetime.time(hour=5),
            lead_time_to_start=pd.Timedelta('1h'),
            run_length=pd.Timedelta('1h')
        )
    )

    proc_fx_obs = create_processed_fxobs(fxobs, fx_series, obs_series)

    res = calculator.calculate_deterministic_metrics(proc_fx_obs, [category],
                                                     ['mbe'])
    assert {(v.index, v.value) for v in res.values} == result


@pytest.mark.parametrize('prob_fx_df,axis,exp_fx_fx_prob', [
    (pd.DataFrame({'25': [1.]*5, '50': [2.]*5, '75': [3.]*5}),
     'y',
     [(pd.Series([1.]*5, name='25'), pd.Series([25.]*5, name='25')),
      (pd.Series([2.]*5, name='50'), pd.Series([50.]*5, name='50')),
      (pd.Series([3.]*5, name='75'), pd.Series([75.]*5, name='75'))]),
])
def test_transform_prob_forecast_value_and_prob(prob_forecasts,
                                                single_observation,
                                                prob_fx_df, axis,
                                                exp_fx_fx_prob,
                                                copy_prob_forecast_with_axis,
                                                create_processed_fxobs):
    conv_prob_fx = copy_prob_forecast_with_axis(prob_forecasts, axis)
    fxobs = datamodel.ForecastObservation(conv_prob_fx, single_observation)
    proc_fxobs = create_processed_fxobs(fxobs, prob_fx_df,
                                        pd.Series([0.]*prob_fx_df.shape[0]))
    result = calculator._transform_prob_forecast_value_and_prob(proc_fxobs)
    assert len(result) == len(exp_fx_fx_prob)
    for res, exp in zip(result, exp_fx_fx_prob):
        assert len(res) == len(exp)
        pd.testing.assert_series_equal(res[0], exp[0])
        pd.testing.assert_series_equal(res[1], exp[1])


@pytest.mark.parametrize('obs,fx_fx_prob,ref_fx_fx_prob,exp_df', [
    (pd.Series([0.]*5),
     [(pd.Series([1.]*5, name='25'), pd.Series([25.]*5, name='25')),
      (pd.Series([2.]*5, name='50'), pd.Series([50.]*5, name='50')),
      (pd.Series([3.]*5, name='75'), pd.Series([75.]*5, name='75'))],
     [(pd.Series([4.]*5, name='25'), pd.Series([25.]*5, name='25')),
      (pd.Series([5.]*5, name='50'), pd.Series([50.]*5, name='50')),
      (pd.Series([6.]*5, name='75'), pd.Series([75.]*5, name='75'))],
     pd.DataFrame({
        (None, 'observation'): pd.Series([0.]*5),
        ('25', 'forecast'): pd.Series([1.]*5),
        ('25', 'probability'): pd.Series([25.]*5),
        ('25', 'reference_forecast'): pd.Series([4.]*5),
        ('25', 'reference_probability'): pd.Series([25.]*5),
        ('50', 'forecast'): pd.Series([2.]*5),
        ('50', 'probability'): pd.Series([50.]*5),
        ('50', 'reference_forecast'): pd.Series([5.]*5),
        ('50', 'reference_probability'): pd.Series([50.]*5),
        ('75', 'forecast'): pd.Series([3.]*5),
        ('75', 'probability'): pd.Series([75.]*5),
        ('75', 'reference_forecast'): pd.Series([6.]*5),
        ('75', 'reference_probability'): pd.Series([75.]*5)}))
])
def test_create_prob_dataframe(obs, fx_fx_prob, ref_fx_fx_prob, exp_df):
    result = calculator._create_prob_dataframe(obs, fx_fx_prob, ref_fx_fx_prob)
    pd.testing.assert_frame_equal(result, exp_df)


@pytest.mark.parametrize('categories', [
    [],
    *list(itertools.combinations(LIST_OF_CATEGORIES, 1)),
    LIST_OF_CATEGORIES[0:1],
    LIST_OF_CATEGORIES[0:2],
    LIST_OF_CATEGORIES,
])
@pytest.mark.parametrize('metrics', [
    *list(itertools.combinations(EVENT_METRICS, 1)),
    EVENT_METRICS[0:1],
    EVENT_METRICS[0:2],
    EVENT_METRICS,
    pytest.param([],
                 marks=pytest.mark.xfail(raises=RuntimeError, strict=True)),
])
def test_calculate_event_metrics(single_event_forecast_observation, categories,
                                 metrics):

    index = pd.DatetimeIndex(
        ["20200301T0000Z", "20200301T0100Z", "20200301T0200Z"]
    )
    obs_values = np.random.randint(0, 2, size=len(index), dtype=bool)
    fx_values = np.random.randint(0, 2, size=len(index), dtype=bool)
    obs_series = pd.Series(obs_values, index=index)
    fx_series = pd.Series(fx_values, index=index)

    fxobs = single_event_forecast_observation
    proc_fx_obs = datamodel.ProcessedForecastObservation(
        name=fxobs.forecast.name,
        original=fxobs,
        interval_value_type=fxobs.forecast.interval_value_type,
        interval_length=fxobs.forecast.interval_length,
        interval_label=fxobs.forecast.interval_label,
        valid_point_count=len(fx_series),
        forecast_values=fx_series,
        observation_values=obs_series,
    )

    result = calculator.calculate_event_metrics(
        proc_fx_obs, categories, metrics
    )

    # Check results
    assert isinstance(result, datamodel.MetricResult)
    assert result.forecast_id == proc_fx_obs.original.forecast.forecast_id
    assert result.name == proc_fx_obs.original.forecast.name
    assert len(result.values) % len(metrics) == 0
    cats = {val.category for val in result.values}
    assert cats == set(categories)
    fx_values = proc_fx_obs.forecast_values
    for cat in categories:
        cat_grps = {v.index for v in result.values if v.category == cat}
        assert len(
            {v.metric for v in result.values if v.category == cat}
        ) == len(metrics)

        # has expected groupings
        if cat == 'month':
            grps = fx_values.groupby(
                fx_values.index.month).groups
            grps = [calendar.month_abbr[g] for g in grps]
        elif cat == 'hour':
            grps = fx_values.groupby(
                fx_values.index.hour).groups
        elif cat == 'year':
            grps = fx_values.groupby(
                fx_values.index.year).groups
        elif cat == 'date':
            grps = fx_values.groupby(
                fx_values.index.date).groups
        elif cat == 'weekday':
            grps = fx_values.groupby(
                fx_values.index.weekday).groups
            grps = [calendar.day_abbr[g] for g in grps]
        elif cat == 'total':
            grps = ['0']
        assert {str(g) for g in grps} == cat_grps
    for val in result.values:
        assert (
            np.isnan(val.value) or
            np.issubdtype(type(val.value), np.number)
        )


@pytest.mark.parametrize('obs_values,fx_values', [
    ([], [True, False]),
    ([False, True], []),
])
def test_calculate_event_metrics_no_data(single_event_forecast_observation,
                                         obs_values, fx_values):

    categories = LIST_OF_CATEGORIES
    metrics = EVENT_METRICS
    obs_series = pd.Series(obs_values, dtype=bool)
    fx_series = pd.Series(fx_values, dtype=bool)

    # processed fx-obs pair
    fxobs = single_event_forecast_observation
    proc_fx_obs = datamodel.ProcessedForecastObservation(
        name=fxobs.forecast.name,
        original=fxobs,
        interval_value_type=fxobs.forecast.interval_value_type,
        interval_length=fxobs.forecast.interval_length,
        interval_label=fxobs.forecast.interval_label,
        valid_point_count=len(fx_series),
        forecast_values=fx_series,
        observation_values=obs_series,
    )

    with pytest.raises(RuntimeError):
        calculator.calculate_event_metrics(
            proc_fx_obs, categories, metrics
        )


@pytest.mark.parametrize('categories', [
    [],
    *list(itertools.combinations(LIST_OF_CATEGORIES, 1)),
    LIST_OF_CATEGORIES[0:1],
    LIST_OF_CATEGORIES[0:2],
    LIST_OF_CATEGORIES,
])
@pytest.mark.parametrize('metrics', [
    *list(itertools.combinations(EVENT_METRICS, 1)),
    EVENT_METRICS[0:1],
    EVENT_METRICS[0:2],
    EVENT_METRICS,
])
def test_calculate_metrics_with_event(single_event_forecast_observation,
                                      categories, metrics):
    index = pd.DatetimeIndex(
        ["20200301T0000Z", "20200301T0100Z", "20200301T0200Z"]
    )

    fxobs = single_event_forecast_observation

    # processed fx-obs pairs
    proc_fx_obs = []
    for i in range(4):
        obs_values = np.random.randint(0, 2, size=len(index), dtype=bool)
        fx_values = np.random.randint(0, 2, size=len(index), dtype=bool)
        obs_series = pd.Series(obs_values, index=index)
        fx_series = pd.Series(fx_values, index=index)

        proc_fx_obs.append(
            datamodel.ProcessedForecastObservation(
                name=fxobs.forecast.name,
                original=fxobs,
                interval_value_type=fxobs.forecast.interval_value_type,
                interval_length=fxobs.forecast.interval_length,
                interval_label=fxobs.forecast.interval_label,
                valid_point_count=len(fx_series),
                forecast_values=fx_series,
                observation_values=obs_series,
            )
        )

    # compute metrics
    result = calculator.calculate_metrics(
        proc_fx_obs, categories, metrics
    )

    assert isinstance(result, list)
    assert isinstance(result[0], datamodel.MetricResult)
    assert len(result) == len(proc_fx_obs)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
@pytest.mark.parametrize("categories,metrics", [
    (LIST_OF_CATEGORIES, EVENT_METRICS),
])
def test_calculate_metrics_with_event_empty(single_event_forecast_observation,
                                            categories, metrics, caplog):
    index = pd.DatetimeIndex(
        ["20200301T0000Z", "20200301T0100Z", "20200301T0200Z"]
    )

    fxobs = single_event_forecast_observation

    # processed fx-obs pairs
    proc_fx_obs = []
    obs_values = np.random.randint(0, 2, size=len(index), dtype=bool)
    obs_series = pd.Series(obs_values, index=index)
    fx_series = pd.Series(dtype=bool)

    proc_fx_obs.append(
        datamodel.ProcessedForecastObservation(
            name=fxobs.forecast.name,
            original=fxobs,
            interval_value_type=fxobs.forecast.interval_value_type,
            interval_length=fxobs.forecast.interval_length,
            interval_label=fxobs.forecast.interval_label,
            valid_point_count=len(fx_series),
            forecast_values=fx_series,
            observation_values=obs_series,
        )
    )

    result = calculator.calculate_metrics(proc_fx_obs, categories, metrics)
    assert len(result) == 0
    assert "ERROR" == caplog.text[0:5]
    failure_log_text = caplog.text[re.search(r'.py:\d+ ', caplog.text).end():]
    assert (f"Failed to calculate event metrics for {proc_fx_obs[0].name}: "
            "No Forecast timeseries data.\n") == failure_log_text



"""A CLI tool for creating reference sites and observations and updating them
with data from their respective API.

As new networks are added, their name should be added as it appears in
`sfa_reference_data.csv` to `NETWORK_OPTIONS`. A module should be created
to handle observation initialization and data update. Each of these network
specific modules should implement two functions:

 * ``initialize_site_observations(api, site)``
    * Where api is an `io.api.APISession` and site is a `datamodel.Site`.
 * ``update_observation_data(api, sites, observations, start, end)``
    * Where api is an `io.api.APISession`, sites is the result of
      `APISession.list_sites`, observations is the result of
      `APISession.list_observations``start` and `end` are datetime
      objects.

The module should then be imported and added to `NETWORKHANDLER_MAP` below,
so that it may be selected based on command line arguments. See the existing
mappings for an example.
"""

import json
import logging
from pkg_resources import resource_filename, Requirement


from requests.exceptions import HTTPError


from solarforecastarbiter.datamodel import Site
from solarforecastarbiter.io.api import APISession
from solarforecastarbiter.io.reference_observations import (
    surfrad,
    solrad,
    crn,
    midc,
    srml,
    rtc,
    common,
    arm,
    pvdaq,
    eia
)


# maps network names to the modules that interact with their api
NETWORKHANDLER_MAP = {
    'NOAA SURFRAD': surfrad,
    'NOAA SOLRAD': solrad,
    'NOAA USCRN': crn,
    'UO SRML': srml,
    'NREL MIDC': midc,
    'DOE RTC': rtc,
    'DOE ARM': arm,
    'NREL PVDAQ': pvdaq,
    'EIA': eia,
}

# list of options for the 'network' argument
NETWORK_OPTIONS = ['NOAA SURFRAD', 'NOAA SOLRAD', 'NOAA USCRN', 'NREL MIDC',
                   'UO SRML', 'DOE RTC', 'DOE ARM', 'NREL PVDAQ', 'EIA']

DEFAULT_SITEFILE = resource_filename(
    Requirement.parse('solarforecastarbiter'),
    'solarforecastarbiter/io/reference_observations/sfa_reference_sites.csv')


logger = logging.getLogger('reference_data')

CLI_DESCRIPTION = """
Tool for initializing and updating SolarForecastArbiter reference observation data.
Supports importing sites from the following networks:

NOAA (The National Oceanic and Atmospheric Administration)

    SURFRAD: Surface Radiation Budget Network
    https://www.esrl.noaa.gov/gmd/grad/surfrad/

    SOLRAD:
    https://www.esrl.noaa.gov/gmd/grad/solrad/index.html

    CRN: U.S. Climate Reference Network
    https://www.ncdc.noaa.gov/crn/

NREL MIDC: National Renewable Energy Laboratory Measurement and Instrumentation Data Center
https://midcdmz.nrel.gov/

UO SRML: University of Oregon Solar Radiation Monitoring Laboratory
http://solardat.uoregon.edu/

DOE RTC: DOE Regional Test Centers for Solar Technologies
https://pv-dashboard.sandia.gov/

DOE ARM: DOE Atmospheric Radiation Measurement
https://www.arm.gov/

NREL PVDAQ: National Renewable Energy Laboratory PV Data Acquisition
https://developer.nrel.gov/docs/solar/pvdaq-v3/

EIA: U.S. Energy Information Administration Open Data
https://www.eia.gov/opendata/
"""  # noqa: E501


def get_apisession(token, base_url=None):
    return APISession(token, base_url=base_url)


def create_site(api, site):
    """Post a new site to the API and create it's applicable reference
    observations.

    Parameters
    ----------
    api : io.APISession
        An APISession with a valid JWT for accessing the Reference Data user.
    site : dict
        Dictionary describing the site to post. This will be instantiated as
        a datamodel.Site object and the value of 'extra_parameters' will be
        serialized to json.

    Returns
    -------
    datamodel.Site
        The created site object.
    """
    # get a reference to network before we serialize extra_parameters
    network = site['extra_parameters']['network']
    network_handler = NETWORKHANDLER_MAP.get(network)
    if network_handler is None:
        logger.warning(f'Unrecognized network, {network} on Site '
                       f'{site["name"]} Observations cannot be '
                       'automatically generated.')
        return
    site.update({'extra_parameters': json.dumps(site['extra_parameters'])})
    site_name = f"{network} {common.clean_name(site['name'])}"
    existing = common.existing_sites(api)
    if site_name in existing:
        logger.info('Site, %s, already exists', site_name)
        created = existing[site_name]
    else:
        site['name'] = site_name
        site_to_create = Site.from_dict(site)
        try:
            created = api.create_site(site_to_create)
        except HTTPError as e:
            logger.error(f"Failed to create Site {site['name']}.")
            logger.debug(f'HTTP Error: {e.response.text}')
            return False
        else:
            logger.info(f'Created Site {created.name} successfully.')
    network_handler.initialize_site_observations(api, created)
    try:
        network_handler.initialize_site_forecasts(api, created)
    except ValueError as e:
        logger.error('Cannot create forecasts for %s: %s', site_name, e)
    return created


def initialize_reference_metadata_objects(token, sites, base_url=None):
    """Instantiate an API session and create reference sites and
    observations.

    Parameters
    ----------
    token : str
        Access token for the SFA API
    sites : list
        List of site dictionary objects. The 'extra_parameters'
        key should contain a nested dict with the following keys:
            network
            network_api_id
            network_api_abbreviation
            observation_interval_length
    base_url : str
        The alternate base url of the SFA API
    """
    logger.info('Initializing reference metadata...')
    api = get_apisession(token, base_url)
    sites_created = 0
    failures = 0
    for site in sites:
        if create_site(api, site):
            sites_created = sites_created + 1
        else:
            failures = failures + 1
    logger.info(f'Created {sites_created} sites successfully, with '
                f'{failures} failures.')


def update_reference_observations(token, start, end, networks, base_url=None):
    """Coordinate updating all existing reference observations.

    Parameters
    ----------
    token : str
        Access token for the SFA API
    start : datetime-like
    end : datetime-like
    networks : list
        List of network names to update.
    base_url : str
        The alternate base url of the SFA API
    """
    api = get_apisession(token, base_url)
    observations = api.list_observations()
    sites = {obs.site for obs in observations}
    for network in networks:
        network_handler = NETWORKHANDLER_MAP.get(network)
        if network_handler is None:
            logger.info(f'{network} observation updates not configured.')
        else:
            network_handler.update_observation_data(api, sites, observations,
                                                    start, end)


def site_df_to_dicts(site_df):
    """ Creates a list of site dictionaries ready to be serialized
    and posted to the api.

    Parameters
    ----------
    site_df: DataFrame
        Pandas Dataframe with the columns:
        interval_length, name, latitude, longitude, elevation,
        timezone, network, network_api_id, network_api_abbreviation,
        attribution

    Returns
    -------
    list
        A list of Site dictionaries.
    """
    site_list = []
    for i, row in site_df.iterrows():
        row = row.fillna('')
        site = {
            'name': row['name'],
            'latitude': row['latitude'],
            'longitude': row['longitude'],
            'elevation': row['elevation'],
            'timezone': row['timezone'],
            'extra_parameters': {
                'network': row['network'],
                'network_api_id': row['network_api_id'],
                'network_api_abbreviation': row['network_api_abbreviation'],
                'observation_interval_length': row['interval_length'],
                'attribution': row['attribution']
            }
        }
        network_handler = NETWORKHANDLER_MAP.get(row['network'])
        if hasattr(network_handler, 'adjust_site_parameters'):
            site = network_handler.adjust_site_parameters(site)
        site_list.append(site)
    return site_list



"""Event forecast error metrics."""

import numpy as np


def _event2count(obs, fx):
    """Convert events (True/False) into counts.

    Given forecasts and observations of events (True=event occurred,
    False=event did not occur), the pairs of forecasts and observations can be
    placed into four categories:
    - True Positive (TP): forecast = event, observed = event
    - False Positive (FP): forecast = event, observed = no event
    - True Negative (TN): forecast = no event, observed = no event
    - False Negative (FN): forecast = no event, observed = event

    Parameters
    ----------
    obs : (n,) array-like
        Observed event values (True=event, False=no event).
    fx : (n,) array-like
        Forecasted event values (True=event, False=no event).

    Returns
    -------
    tp : int
        Number of true positives.
    fp : int
        Number of false positives.
    tn : int
        Number of true negatives.
    fn : int
        Number of false negatives.

    Raises
    ------
    RuntimeError
        If there is no forecast or observation timeseries data, or the forecast
        and observation timeseries data do not have the same length.

    """

    obs = np.asarray(obs)
    fx = np.asarray(fx)

    if len(obs) == 0:
        raise RuntimeError("No Observation timeseries data.")
    elif len(fx) == 0:
        raise RuntimeError("No Forecast timeseries data.")
    elif len(obs) != len(fx):
        raise RuntimeError("Forecast and Observation timeseries data do not "
                           "have the same length.")

    tp = np.count_nonzero(np.logical_and(fx, obs))
    fp = np.count_nonzero(np.logical_and(fx, ~obs))
    tn = np.count_nonzero(np.logical_and(~fx, ~obs))
    fn = np.count_nonzero(np.logical_and(~fx, obs))
    return tp, fp, tn, fn


def probability_of_detection(obs, fx):
    """Probability of Detection (POD).

    .. math:: \\text{POD} = \\text{TP} / (\\text{TP} + \\text{FN})

    Parameters
    ----------
    obs : (n,) array-like
        Observed event values (True=event, False=no event).
    fx : (n,) array-like
        Forecasted event values (True=event, False=no event).

    Returns
    -------
    pod : float
        The POD of the forecast.

    Raises
    ------
    RuntimeError
        If there is no forecast or observation timeseries data, or the forecast
        and observation timeseries data do not have the same length.

    """

    tp, fp, tn, fn = _event2count(obs, fx)
    if (tp + fn) == 0:
        return 0.0
    else:
        return tp / (tp + fn)


def false_alarm_ratio(obs, fx):
    """False Alarm Ratio (FAR).

    .. math:: \\text{FAR} = \\text{FP} / (\\text{TP} + \\text{FP})

    Parameters
    ----------
    obs : (n,) array-like
        Observed event values (True=event, False=no event).
    fx : (n,) array-like
        Forecasted event values (True=event, False=no event).

    Returns
    -------
    far : float
        The FAR of the forecast.

    Raises
    ------
    RuntimeError
        If there is no forecast or observation timeseries data, or the forecast
        and observation timeseries data do not have the same length.

    """
    tp, fp, tn, fn = _event2count(obs, fx)
    if (tp + fp) == 0:
        return 0.0
    else:
        return fp / (tp + fp)


def probability_of_false_detection(obs, fx):
    """Probability of False Detection (POFD).

    .. math:: \\text{POFD} = \\text{FP} / (\\text{FP} + \\text{TN})

    Parameters
    ----------
    obs : (n,) array-like
        Observed event values (True=event, False=no event).
    fx : (n,) array-like
        Forecasted event values (True=event, False=no event).

    Returns
    -------
    pofd : float
        The POFD of the forecast.

    Raises
    ------
    RuntimeError
        If there is no forecast or observation timeseries data, or the forecast
        and observation timeseries data do not have the same length.

    """

    tp, fp, tn, fn = _event2count(obs, fx)
    if (fp + tn) == 0:
        return 0.0
    else:
        return fp / (fp + tn)


def critical_success_index(obs, fx):
    """Critical Success Index (CSI).

    .. math:: \\text{CSI} = \\text{TP} / (\\text{TP} + \\text{FP} + \\text{FN})

    Parameters
    ----------
    obs : (n,) array-like
        Observed event values (True=event, False=no event).
    fx : (n,) array-like
        Forecasted event values (True=event, False=no event).

    Returns
    -------
    csi : float
        The CSI of the forecast.

    Raises
    ------
    RuntimeError
        If there is no forecast or observation timeseries data, or the forecast
        and observation timeseries data do not have the same length.

    """

    tp, fp, tn, fn = _event2count(obs, fx)
    if (tp + fp + fn) == 0:
        return 0.0
    else:
        return tp / (tp + fp + fn)


def event_bias(obs, fx):
    """Event Bias (EBIAS).

    .. math:: \\text{EBIAS} = (\\text{TP} + \\text{FP}) / (\\text{TP} + \\text{FN})  # NOQA

    Parameters
    ----------
    obs : (n,) array-like
        Observed event values (True=event, False=no event).
    fx : (n,) array-like
        Forecasted event values (True=event, False=no event).

    Returns
    -------
    ebias : float
        The EBIAS of the forecast.

    Raises
    ------
    RuntimeError
        If there is no forecast or observation timeseries data, or the forecast
        and observation timeseries data do not have the same length.

    """

    tp, fp, tn, fn = _event2count(obs, fx)
    if (tp + fn) == 0:
        return 0.0
    else:
        return (tp + fp) / (tp + fn)


def event_accuracy(obs, fx):
    """Event Accuracy (EA).

    .. math:: \\text{EA} = (\\text{TP} + \\text{TN}) / n

    where n is the number of samples.

    Parameters
    ----------
    obs : (n,) array-like
        Observed event values (True=event, False=no event).
    fx : (n,) array-like
        Forecasted event values (True=event, False=no event).

    Returns
    -------
    ea : float
        The EA of the forecast.

    Raises
    ------
    RuntimeError
        If there is no forecast or observation timeseries data, or the forecast
        and observation timeseries data do not have the same length.

    """

    n = len(obs)
    tp, fp, tn, fn = _event2count(obs, fx)
    return (tp + tn) / n


# Add new metrics to this map to map shorthand to function
_MAP = {
    'pod': (probability_of_detection, 'POD'),
    'far': (false_alarm_ratio, 'FAR'),
    'pofd': (probability_of_false_detection, 'POFD'),
    'csi': (critical_success_index, 'CSI'),
    'ebias': (event_bias, 'EBIAS'),
    'ea': (event_accuracy, 'EA')
}

__all__ = [m[0].__name__ for m in _MAP.values()]

# Functions that require a reference forecast
_REQ_REF_FX = []

# Functions that require normalized factor
_REQ_NORM = []



"""Functions to extract EIA."""

import requests
import pandas as pd


def get_eia_data(series_id, api_key, start, end):
    """
    Retrieve data from the EIA Open Data API. [1]_

    Parameters
    ----------
    series_id : string
        The series ID in the EIA API.
    api_key : string
        The API key for accessing the EIA API.
    start : pd.Timestamp
        The start timestamp in UTC.
    end : pd.Timestamp
        The end timestamp in UTC.

    Returns
    -------
    pandas.DataFrame
        The data from *start* to *end* for *series_id*.

    References
    ----------
    .. [1] https://www.eia.gov/opendata/

    Examples
    --------
    >>> series_id = "EBA.CISO-ALL.D.H"  # CAISO demand [MW]
    >>> start = pd.Timestamp("20200601T0800Z")
    >>> end = pd.Timestamp("20200602T0800Z")
    >>> api_key = "yourapikeygoeshere"
    >>> get_eia_data(series_id, api_key, start, end)
    timestamp                    value
    2020-06-01 08:00:00+00:00    22028
    2020-06-01 09:00:00+00:00    21141
    2020-06-01 10:00:00+00:00    20573
    ...

    """

    base_url = "https://api.eia.gov/series/"
    params = dict(
        api_key=api_key,
        series_id=series_id,
        start=start.strftime("%Y%m%dT%HZ"),
        end=end.strftime("%Y%m%dT%HZ"),
    )
    r = requests.get(base_url, params=params)
    r.raise_for_status()

    data = r.json()["series"][0]["data"]
    df = pd.DataFrame(data, columns=["timestamp", "value"])
    df["timestamp"] = pd.to_datetime(df["timestamp"])
    df.set_index("timestamp", inplace=True)
    df.sort_index(inplace=True)

    # replace invalid values with NaN
    # - "null" if missing
    # - "w" if witheld
    # - "*" if statistically insignificant
    df["value"] = pd.to_numeric(df["value"], errors="coerce")

    return df



import datetime as dt
import pandas as pd
import pytest
from urllib import error

from pandas import Timestamp
from pandas.testing import assert_frame_equal

from solarforecastarbiter.io.reference_observations import srml
from solarforecastarbiter.datamodel import Site


test_site_dict = {
    "elevation": 595.0,
    "extra_parameters": "{\"network_api_id\": \"94040.0\", \"attribution\": \"Peterson, J., and Vignola, F., 2017: Structure of a Comprehensive Solar Radiation Dataset. Proceedings of the ASES National Solar Conference 2017. doi: 10.18086/solar.2017.07.02\", \"network\": \"UO SRML\", \"network_api_abbreviation\": \"AS\", \"observation_interval_length\": 1.0}",  # noqa
    "latitude": 42.19,
    "longitude": -122.7,
    "modeling_parameters": {
        "ac_capacity": 0.02,
        "ac_loss_factor": 0.0,
        "dc_capacity": 0.02,
        "dc_loss_factor": 0.0,
        "surface_azimuth": 180.0,
        "surface_tilt": 15.0,
        "temperature_coefficient": 0.3,
        "tracking_type": "fixed"
    },
    "name": "Ashland OR PV",
    "timezone": "Etc/GMT+8",
    "provider": "",
    "site_id": "",
}


test_site_object = Site.from_dict(test_site_dict)


@pytest.fixture()
def test_site():
    return test_site_object


srml_df = pd.DataFrame(
    {'ghi_1': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 0.0},
     'ghi_1_flag': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11},
     'dni_1': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 0.0},
     'dni_1_flag': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11},
     'dhi_1': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 0.0},
     'dhi_1_flag': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11},
     '1161': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 0.0},
     '1161_flag': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11},
     '5161': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 235.0},
     '5161_flag': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11},
     '5162': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 250.0},
     '5162_flag': {Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11},
     'wind_speed_3': {
         Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 0.1},
     'wind_speed_3_flag': {
         Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11},
     'temp_air_1': {
         Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 2.1},
     'temp_air_1_flag': {
         Timestamp('2020-03-01 00:00:00-0800', tz='Etc/GMT+8'): 11}})


@pytest.fixture()
def test_data():
    return srml_df


@pytest.mark.parametrize('start,end,exp', [
    (dt.datetime(2019, 1, 10, tzinfo=dt.timezone.utc),
     dt.datetime(2019, 3, 11, tzinfo=dt.timezone.utc),
     [(2019, 1), (2019, 2), (2019, 3)]),
    (dt.datetime(2019, 11, 10, tzinfo=dt.timezone.utc),
     dt.datetime(2020, 2, 11, tzinfo=dt.timezone.utc),
     [(2019, 11), (2019, 12), (2020, 1), (2020, 2)]),
    (dt.datetime(2019, 11, 10, tzinfo=dt.timezone.utc),
     dt.datetime(2019, 11, 11, tzinfo=dt.timezone.utc),
     [(2019, 11)]),
    (dt.datetime(2019, 11, 10, tzinfo=dt.timezone.utc),
     dt.datetime(2021, 2, 11, tzinfo=dt.timezone.utc),
     [(2019, 11), (2019, 12), (2020, 1), (2020, 2), (2020, 3),
      (2020, 4), (2020, 5), (2020, 6), (2020, 7), (2020, 8),
      (2020, 9), (2020, 10), (2020, 11), (2020, 12), (2021, 1),
      (2021, 2)]),
    (dt.datetime(2019, 10, 1, tzinfo=dt.timezone.utc),
     dt.datetime(2019, 9, 1, tzinfo=dt.timezone.utc), []),
    (dt.datetime(2019, 10, 1, tzinfo=dt.timezone.utc),
     dt.datetime(2018, 11, 1, tzinfo=dt.timezone.utc), [])
])
def test_fetch(mocker, single_site, start, end, exp):
    rd = mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.request_data',
        return_value=None)
    srml.fetch('', single_site, start, end)
    year_month = [(ca[0][1], ca[0][2]) for ca in rd.call_args_list]
    assert year_month == exp


@pytest.mark.parametrize('start,end', [
    (dt.datetime(2019, 1, 10, tzinfo=dt.timezone.utc),
     dt.datetime(2020, 3, 11, 8, 1, tzinfo=dt.timezone.utc)),
])
def test_fetch_power_conversion(
        mocker, single_site, start, end, test_data):
    mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.request_data',
        return_value=test_data)
    data = srml.fetch('', single_site, start, end)
    assert not [col for col in data.columns if '_flag' in col]
    assert data['5161'].iloc[0] == 0.000235
    assert data['5162'].iloc[0] == 0.000250


def test_fetch_tz(single_site):
    start = dt.datetime(2019, 1, 1, tzinfo=dt.timezone.utc)
    with pytest.raises(TypeError):
        srml.fetch('', single_site, start,
                   dt.datetime(2019, 2, 1))


@pytest.mark.parametrize('api_id,has_power', [
    (94040.0, True),
    (94291.0, False),
])
def test_adjust_parameters(api_id, has_power):
    site = srml.adjust_site_parameters({
        'extra_parameters': {
            'network_api_id': str(api_id),
            }
    })
    if has_power:
        modeling_params = site['modeling_parameters']
        assert modeling_params['ac_capacity'] == 0.02
        assert modeling_params['dc_capacity'] == 0.02
        assert modeling_params['ac_loss_factor'] == 0.0
        assert modeling_params['dc_loss_factor'] == 0.0
        assert modeling_params['surface_azimuth'] == 180.0
        assert modeling_params['surface_tilt'] == 15.0
        assert modeling_params['temperature_coefficient'] == 0.3
        assert modeling_params['tracking_type'] == 'fixed'
    else:
        assert 'modeling_parameters' not in site


def test_init_site_observations(
        mocker, test_data, test_site):
    mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.request_data',
        return_value=test_data)
    mock_create_obs = mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.common.'
        'create_observation')
    mock_chk_post = mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.common.'
        'check_and_post_observation')
    mock_api = mocker.MagicMock()
    srml.initialize_site_observations(mock_api, test_site)
    assert mock_create_obs.call_count == 5
    assert mock_chk_post.call_count == 2


def test_request_data(mocker, test_site, test_data):
    mocked_iotools = mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.iotools')
    mocked_iotools.read_srml_month_from_solardat = mocker.MagicMock(
        return_value=test_data)
    data = srml.request_data(test_site, 1, 1)
    assert_frame_equal(data, test_data)


@pytest.mark.parametrize('exception', [
    pd.errors.EmptyDataError,
    error.URLError,
])
def test_request_data_warnings(mocker, exception, test_site):
    mocked_iotools = mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.iotools')
    mocked_iotools.read_srml_month_from_solardat = mocker.MagicMock(
        side_effect=exception('error'))
    logger = mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.logger')
    data = srml.request_data(test_site, 1, 1)
    assert logger.warning.call_count == 3
    assert data is None


def test_initialize_site_forecasts(mocker, test_site):
    mock_create_fx = mocker.patch(
        'solarforecastarbiter.io.reference_observations.srml.common.'
        'create_forecasts')
    mock_api = mocker.MagicMock()
    srml.initialize_site_forecasts(mock_api, test_site)
    assert 'ac_power' in mock_create_fx.call_args[0][2]
    assert 'dc_power' in mock_create_fx.call_args[0][2]

    regular_site_dict = test_site_dict.copy()
    regular_site_dict.pop('modeling_parameters')
    reg_site = Site.from_dict(regular_site_dict)
    srml.initialize_site_forecasts(mock_api, reg_site)
    assert 'ac_power' not in mock_create_fx.call_args[0][2]
    assert 'dc_power' not in mock_create_fx.call_args[0][2]



import datetime as dt
from functools import wraps
import logging


from bokeh.embed import components
from bokeh.layouts import gridplot
from bokeh.models import ColumnDataSource, Label, HoverTool
from bokeh.plotting import figure
from bokeh import palettes
from matplotlib import cm
from matplotlib.colors import Normalize
import plotly.graph_objects as go
import pandas as pd
import pytz


from solarforecastarbiter.plotting import utils as plot_utils
from solarforecastarbiter.validation import quality_mapping


UTCS = (dt.timezone.utc, pytz.UTC)
logger = logging.getLogger('sfa.plotting.timeseries')
PLOT_WIDTH = 900
PALETTE = palettes.all_palettes['Category20b'][20][::4]
# flags with color None will be assigned a color from PALETTE
FLAG_COLORS = {
    'MISSING': '#e6550d',
    'NOT VALIDATED': '#ff7f0e',
    'USER FLAGGED': '#d62728',
    'NIGHTTIME': None,
    'DAYTIME': None,
    'CLEARSKY': None,
    'SHADED': None,
    'UNEVEN FREQUENCY': None,
    'LIMITS EXCEEDED': None,
    'CLEARSKY EXCEEDED': None,
    'STALE VALUES': None,
    'DAYTIME STALE VALUES': None,
    'INTERPOLATED VALUES': None,
    'DAYTIME INTERPOLATED VALUES': None,
    'CLIPPED VALUES': None,
    'INCONSISTENT IRRADIANCE COMPONENTS': None,
}


def build_figure_title(object_name, start, end):
    """Builds a title for the plot

    Parameters
    ----------
    object_name : str
        Name of the object being plotted

    start: datetime-like
        The start of the interval being plotted.

    end: datetime-like
        The end of the interval being plotted.

    Returns
    -------
    string
        The appropriate figure title.

    Raises
    ------
    ValueError
        If start or end is not localized to UTC
    """
    if start.tzinfo not in UTCS or end.tzinfo not in UTCS:
        raise ValueError('start and end must be localized to UTC')

    start_string = start.strftime('%Y-%m-%d %H:%M')
    end_string = end.strftime('%Y-%m-%d %H:%M')
    figure_title = (f'{object_name} {start_string} to {end_string} UTC')
    return figure_title


def _single_quality_bar(flag_name, plot_width, x_range, color, source):
    qfig = figure(sizing_mode='stretch_width',
                  plot_height=30,
                  plot_width=plot_width,
                  x_range=x_range,
                  toolbar_location=None,
                  min_border_bottom=0,
                  min_border_top=0,
                  tools='xpan',
                  x_axis_location=None,
                  y_axis_location=None)
    qfig.ygrid.grid_line_color = None
    qfig.line(x='timestamp', y=flag_name,
              line_width=qfig.plot_height,
              source=source, alpha=0.6,
              line_color=color)
    flag_label = Label(x=5, y=0,
                       x_units='screen', y_units='screen',
                       text=flag_name, render_mode='css',
                       border_line_color=None,
                       background_fill_alpha=0,
                       text_font_size='1em',
                       text_font_style='bold')
    qfig.add_layout(flag_label)
    return qfig


def make_quality_bars(source, plot_width, x_range):
    """
    Make figures to display the whether a time is flagged for any
    of the columns in source.

    Parameters
    ----------
    source : bokeh.models.ColumnDataSource
        The predefined data source with flags loaded. Only columns
        in FLAG_COLORS will be made into bars. If data for a flag
        is empty, a bar will not be generated for that flag.
    plot_width : int
        The width of the figures
    x_range : bokeh.Range or tuple
        If x_range is a bokeh Range from another plot, the plots will
        be linked on panning/zooming/etc.

    Returns
    -------
    list
       Of bar figures. The top figure will have an appropriate title.
    """
    palette = iter(PALETTE * 3)
    out = []
    for flag, color in FLAG_COLORS.items():
        # only display bars for flags that have at least on occurence
        if flag not in source.data or pd.isna(source.data[flag]).all():
            continue
        if color is None:
            color = next(palette)
        nextfig = _single_quality_bar(flag, plot_width, x_range,
                                      color, source)
        out.append(nextfig)

    # add the title to the top bar if there are any bars
    if out:
        out[0].plot_height = 60
        out[0].title.text = 'Quality Flags'
        out[0].title.text_font_size = '1em'

    return out


def add_hover_tool(fig, source, **hover_kwargs):
    """Add a hover tool to fig. If `add_line=True` in `hover_kwargs`
    an invisible line is added to enable hover values step plots"""
    if hover_kwargs.pop('add_line', False):
        fig.line(x='timestamp', y='value', source=source, line_alpha=0)

    tooltips = [
        ('timestamp', '@timestamp{%FT%H:%M:%S%z}'),
        ('value', '@value{%0.2f}')
    ]
    if 'active_flags' in source.data.keys():
        tooltips.append(('quality flags', '@active_flags'))
    hover = HoverTool(tooltips=tooltips, formatters={
        'timestamp': 'datetime', 'value': 'printf'}, mode='vline',
                      **hover_kwargs)
    fig.add_tools(hover)


def make_basic_timeseries(source, object_name, variable, interval_label,
                          plot_width):
    """
    Make a basic timeseries plot (with either a step or line)
    and add a hover tool.

    Parameters
    ----------
    source : bokeh.models.ColumnDataSource
        The datasource with 'timestamp' and 'value' columns that will be
        plotted.
    object_name : str
        Name of the object to be plotted so an appropriate title can be made.
    variable : str
        Variable of the plotted object
    interval_label : str
        Interval label of the object to determine whether a line or step
        is most appropriate.
    plot_width : int
        The width of the output figure

    Returns
    -------
    bokeh.models.Figure
        The figure with the timeseries

    Raises
    ------
    KeyError
       If timestamp is not a column in source
    IndexError
       If the timestamp column is empty
    """
    timestamps = source.data['timestamp']
    first = pd.Timestamp(timestamps[0], tz='UTC')
    last = pd.Timestamp(timestamps[-1], tz='UTC')
    plot_method, plot_kwargs, hover_kwargs = plot_utils.line_or_step(
        interval_label)
    figure_title = build_figure_title(object_name, first, last)
    fig = figure(title=figure_title, sizing_mode='scale_width',
                 plot_width=plot_width,
                 plot_height=300, x_range=(first, last),
                 x_axis_type='datetime',
                 tools='pan,wheel_zoom,box_zoom,zoom_in,zoom_out,reset,save',
                 toolbar_location='above',
                 min_border_bottom=50)
    getattr(fig, plot_method)(x='timestamp', y='value', source=source,
                              **plot_kwargs)
    fig.yaxis.axis_label = plot_utils.format_variable_name(variable)
    fig.xaxis.axis_label = 'Time (UTC)'
    if variable == 'event':
        fig.yaxis.ticker = [0, 1]
        fig.yaxis.major_label_overrides = {1: 'True', 0: 'False'}
    add_hover_tool(fig, source, **hover_kwargs)
    return fig


def _make_layout(figs):
    layout = gridplot(figs,
                      ncols=1,
                      merge_tools=True,
                      toolbar_location='above',
                      sizing_mode='scale_width')
    return layout


def to_components(f):
    """Return script and div of a bokeh object if the return_components
    kwarg is True"""
    @wraps(f)
    def wrapper(*args, **kwargs):
        if kwargs.pop('return_components', False):
            out = f(*args, **kwargs)
            if out is not None:
                return components(out)
            else:
                return out
        else:
            return f(*args, **kwargs)
    return wrapper


@to_components
def generate_forecast_figure(forecast, data, limit=None):
    """
    Creates a bokeh timeseries figure for forcast data

    Parameters
    ----------
    forecast : datamodel.Forecast
        The Forecast that is being plotted

    data : pandas.Series
        The forecast data with a datetime index to be plotted

    limit : pandas.Timedelta or None
        The time limit from the last datapoint to plot. If None, all
        data is plotted.


    Returns
    -------
    None
        When the data is empty
    script, div : str
        When return_components = True, return the <script> and <div>
        components for the Bokeh plot.
    bokeh components from gridplot
        When return_components = False
    """
    logger.info('Starting forecast figure generation...')
    if len(data.index) == 0:
        return None
    data = plot_utils.align_index(data, forecast.interval_length, limit)
    cds = ColumnDataSource(data.reset_index())
    fig = make_basic_timeseries(cds, forecast.name, forecast.variable,
                                forecast.interval_label, PLOT_WIDTH)
    layout = _make_layout([fig])
    logger.info('Figure generated succesfully')
    return layout


@to_components
def generate_observation_figure(observation, data, limit=pd.Timedelta('3d')):
    """
    Creates a bokeh figure from API responses for an observation

    Parameters
    ----------
    observation : datamodel.Observation
        The Observation that is being plotted

    data : pandas.DataFrame
        The observation data to be plotted with datetime index
        and ('value', 'quality_flag') columns

    limit : pandas.Timedelta or None
        The time limit from the last datapoint to plot. If None, all
        data is plotted.

    Returns
    -------
    None
        When the data is empty
    script, div : str
        When return_components = True, return the <script> and <div>
        components for the Bokeh plot.
    bokeh components from gridplot
        When return_components = False
    """
    logger.info('Starting observation forecast generation...')
    if len(data.index) == 0:
        return None
    data = plot_utils.align_index(data, observation.interval_length, limit)
    quality_flag = data.pop('quality_flag').dropna().astype(int)
    bool_flags = quality_mapping.convert_mask_into_dataframe(quality_flag)
    active_flags = quality_mapping.convert_flag_frame_to_strings(bool_flags)
    active_flags.name = 'active_flags'
    flags = bool_flags.mask(~bool_flags).reindex(data.index)  # add missing
    flags['MISSING'] = pd.Series(1.0, index=data.index)[pd.isna(data['value'])]
    # need to fill as line needs more than a single point to show up
    if observation.interval_label == 'ending':
        flags.bfill(axis=0, limit=1, inplace=True)
    else:
        # for interval beginning and instantaneous
        flags.ffill(axis=0, limit=1, inplace=True)
    cds = ColumnDataSource(pd.concat([data, flags, active_flags], axis=1))
    figs = [make_basic_timeseries(cds, observation.name, observation.variable,
                                  observation.interval_label, PLOT_WIDTH)]

    figs.extend(make_quality_bars(cds, PLOT_WIDTH, figs[0].x_range))
    layout = _make_layout(figs)
    logger.info('Figure generated succesfully')
    return layout


PLOTLY_MARGINS = {'l': 50, 'r': 50, 'b': 50, 't': 100, 'pad': 4}
PLOTLY_LAYOUT_DEFAULTS = {
    'autosize': True,
    'height': 300,
    'margin': PLOTLY_MARGINS,
    'plot_bgcolor': '#FFF',
    'font': {'size': 14}
}


def _plot_probabilsitic_distribution_axis_y(fig, forecast, data):
    """
    Plot all probabilistic forecast values for axis='y' by adding traces to
    fig.

    Parameters
    ----------
    fig: plotly.graph_objects.Figure
    forecast: :py:class`solarforecastarbiter.datamodel.ProbabilisticForecast`
    data: pd.DataFrame
    """
    color_map = cm.get_cmap('viridis')
    color_scaler = cm.ScalarMappable(
        Normalize(vmin=0, vmax=1),
        color_map,
    )

    units = forecast.units

    percentiles_are_symmetric = plot_utils.percentiles_are_symmetric(
        data.columns.values.astype('float'))

    # may not work for constant values that don't convert nicely from str/float
    constant_values = data.columns.astype('float').sort_values()
    for i, constant_value in enumerate(constant_values):
        if i == 0:
            fill = None
        else:
            fill = 'tonexty'

        if percentiles_are_symmetric:
            if constant_value <= 50 and i != 0:
                fill_value = constant_values[i - 1]
            else:
                fill_value = constant_value
            fill_value = 2 * abs(fill_value - 50)
        else:
            fill_value = 100 - constant_value

        fill_color = plot_utils.distribution_fill_color(
            color_scaler, fill_value)

        plot_kwargs = plot_utils.line_or_step_plotly(forecast.interval_label)

        forecast_name = f'Prob(f <= x) = {str(constant_value)}%'

        go_ = go.Scatter(
            x=data.index,
            y=data[str(constant_value)],
            name=f'{str(constant_value)} %',
            hovertemplate=(
                f'<b>{forecast_name}</b><br>'
                '<b>Value</b>: %{y} '+f'{units}<br>'
                '<b>Time</b>: %{x}<br>'),
            connectgaps=False,
            showlegend=False,
            mode='lines',
            fill=fill,
            fillcolor=fill_color,
            line=dict(
                color=fill_color,
            ),
            **plot_kwargs,
        )
        fig.add_trace(go_)


def _plot_probabilsitic_distribution_axis_x(fig, forecast, data):
    """
    Plot all probabilistic forecast values for axis='x' by adding traces to
    fig.

    Parameters
    ----------
    fig: plotly.graph_objects.Figure
    forecast: :py:class`solarforecastarbiter.datamodel.ProbabilisticForecast`
    data: pd.DataFrame
    """
    palette = iter(PALETTE * 3)

    units = forecast.units

    for constant_value in data.columns:
        line_color = next(palette)

        plot_kwargs = plot_utils.line_or_step_plotly(forecast.interval_label)

        forecast_name = f'Prob(x <= {str(constant_value)} {units})'
        go_ = go.Scatter(
            x=data.index,
            y=data[str(constant_value)],
            name=forecast_name,
            hovertemplate=(
                f'<b>{forecast_name}</b><br>'
                '<b>Value</b>: %{y} %<br>'
                '<b>Time</b>: %{x}<br>'),
            connectgaps=False,
            showlegend=True,
            mode='lines',
            line=dict(
                color=line_color,
            ),
            **plot_kwargs,
        )
        fig.add_trace(go_)


def generate_probabilistic_forecast_figure(
        forecast, data, limit=pd.Timedelta('3d')):
    """
    Creates a plotly figure spec from api response for a probabilistic forecast
    group.

    Parameters
    ----------
    forecast : datamodel.ProbabilisticForecast
    data : pandas.DataFrame
        DataFrame with forecast values in each column, column names as the
        constant values and a datetime index.
    limit : pandas.Timedelta or None

    Returns
    -------
    None
        When the data is empty.
    figure: Plotly.graph_objects.Figure
        Plotly json specification for the plot.
    """
    logger.info('Starting probabilistic forecast figure generation...')
    if len(data.index) == 0:
        return None

    fig = go.Figure()
    if 'x' in forecast.axis:
        ylabel = 'Probability (%)'
        _plot_probabilsitic_distribution_axis_x(fig, forecast, data)
    else:
        ylabel = plot_utils.format_variable_name(forecast.variable)
        _plot_probabilsitic_distribution_axis_y(fig, forecast, data)
    fig.update_xaxes(title_text=f'Time (UTC)', showgrid=True,
                     gridwidth=1, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside')
    fig.update_yaxes(title_text=ylabel, showgrid=True,
                     gridwidth=1, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside',
                     fixedrange=True)
    first = data.index[0]
    last = data.index[-1]
    fig.update_layout(
        title=build_figure_title(forecast.name, first, last),
        legend=dict(font=dict(size=10)),
        **PLOTLY_LAYOUT_DEFAULTS,
    )
    return fig
